@@ -14,16 +14,18 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-from neutron.db import common_db_mixin as base_db
 from neutron.db import models_v2
-from neutron.plugins.common import utils
 from neutron_lib.callbacks import events
 from neutron_lib.callbacks import registry
 from neutron_lib.callbacks import resources
 from neutron_lib import constants as lib_constants
+from neutron_lib.db import api as db_api
+from neutron_lib.db import model_query
+from neutron_lib.db import utils as db_utils
 from neutron_lib.exceptions import l3 as l3_exception
 from neutron_lib.plugins import constants as p_constants
 from neutron_lib.plugins import directory
+from neutron_lib.plugins import utils
 from oslo_log import log as logging
 from oslo_utils import excutils
 from oslo_utils import uuidutils
@@ -41,8 +43,7 @@ LOG = logging.getLogger(__name__)
 
 
 class VPNPluginDb(vpnaas.VPNPluginBase,
-                  vpn_endpoint_groups.VPNEndpointGroupsPluginBase,
-                  base_db.CommonDbMixin):
+                  vpn_endpoint_groups.VPNEndpointGroupsPluginBase):
     """VPN plugin database class using SQLAlchemy models."""
 
     def _get_validator(self):
@@ -57,13 +58,13 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         return vpn_validator.VpnReferenceValidator()
 
     def update_status(self, context, model, v_id, status):
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             v_db = self._get_resource(context, model, v_id)
             v_db.update({'status': status})
 
     def _get_resource(self, context, model, v_id):
         try:
-            r = self._get_by_id(context, model, v_id)
+            r = model_query.get_by_id(context, model, v_id)
         except exc.NoResultFound:
             with excutils.save_and_reraise_exception(reraise=False) as ctx:
                 if issubclass(model, vpn_models.IPsecSiteConnection):
@@ -118,7 +119,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
                'peer_ep_group_id': ipsec_site_conn['peer_ep_group_id'],
                }
 
-        return self._fields(res, fields)
+        return db_utils.resource_fields(res, fields)
 
     def get_endpoint_info(self, context, ipsec_sitecon):
         """Obtain all endpoint info, and store in connection for validation."""
@@ -151,7 +152,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         ipsec_sitecon = ipsec_site_connection['ipsec_site_connection']
         validator = self._get_validator()
         validator.assign_sensible_ipsec_sitecon_defaults(ipsec_sitecon)
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             #Check permissions
             vpnservice_id = ipsec_sitecon['vpnservice_id']
             self._get_resource(context, vpn_models.VPNService, vpnservice_id)
@@ -194,7 +195,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
             for cidr in ipsec_sitecon['peer_cidrs']:
                 peer_cidr_db = vpn_models.IPsecPeerCidr(
                     cidr=cidr,
-                    ipsec_site_connection_id=ipsec_site_conn_db['id']
+                    ipsec_site_connection_id=ipsec_site_conn_db.id
                 )
                 context.session.add(peer_cidr_db)
         return self._make_ipsec_site_connection_dict(ipsec_site_conn_db)
@@ -205,7 +206,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         ipsec_sitecon = ipsec_site_connection['ipsec_site_connection']
         changed_peer_cidrs = False
         validator = self._get_validator()
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             ipsec_site_conn_db = self._get_resource(
                 context, vpn_models.IPsecSiteConnection, ipsec_site_conn_id)
             vpnservice_id = ipsec_site_conn_db['vpnservice_id']
@@ -244,13 +245,13 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
             del ipsec_sitecon["peer_cidrs"]
             if ipsec_sitecon:
                 ipsec_site_conn_db.update(ipsec_sitecon)
-        result = self._make_ipsec_site_connection_dict(ipsec_site_conn_db)
+            result = self._make_ipsec_site_connection_dict(ipsec_site_conn_db)
         if changed_peer_cidrs:
             result['peer_cidrs'] = new_peer_cidrs
         return result
 
     def delete_ipsec_site_connection(self, context, ipsec_site_conn_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             ipsec_site_conn_db = self._get_resource(
                 context, vpn_models.IPsecSiteConnection, ipsec_site_conn_id)
             context.session.delete(ipsec_site_conn_db)
@@ -268,12 +269,13 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
             ipsec_site_conn_db, fields)
 
     def get_ipsec_site_connections(self, context, filters=None, fields=None):
-        return self._get_collection(context, vpn_models.IPsecSiteConnection,
-                                    self._make_ipsec_site_connection_dict,
-                                    filters=filters, fields=fields)
+        return model_query.get_collection(
+            context, vpn_models.IPsecSiteConnection,
+            self._make_ipsec_site_connection_dict,
+            filters=filters, fields=fields)
 
     def update_ipsec_site_conn_status(self, context, conn_id, new_status):
-        with context.session.begin():
+        with db_api.CONTEXT_WRITER.using(context):
             self._update_connection_status(context, conn_id, new_status, True)
 
     def _update_connection_status(self, context, conn_id, new_status,
@@ -307,7 +309,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
                'pfs': ikepolicy['pfs']
                }
 
-        return self._fields(res, fields)
+        return db_utils.resource_fields(res, fields)
 
     def create_ikepolicy(self, context, ikepolicy):
         ike = ikepolicy['ikepolicy']
@@ -316,7 +318,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         lifetime_units = lifetime_info.get('units', 'seconds')
         lifetime_value = lifetime_info.get('value', 3600)
 
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             validator.validate_ike_policy(context, ike)
             ike_db = vpn_models.IKEPolicy(
                 id=uuidutils.generate_uuid(),
@@ -338,7 +340,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
     def update_ikepolicy(self, context, ikepolicy_id, ikepolicy):
         ike = ikepolicy['ikepolicy']
         validator = self._get_validator()
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             validator.validate_ike_policy(context, ike)
             if context.session.query(vpn_models.IPsecSiteConnection).filter_by(
                     ikepolicy_id=ikepolicy_id).first():
@@ -356,7 +358,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         return self._make_ikepolicy_dict(ike_db)
 
     def delete_ikepolicy(self, context, ikepolicy_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             if context.session.query(vpn_models.IPsecSiteConnection).filter_by(
                     ikepolicy_id=ikepolicy_id).first():
                 raise vpnaas.IKEPolicyInUse(ikepolicy_id=ikepolicy_id)
@@ -370,9 +372,9 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         return self._make_ikepolicy_dict(ike_db, fields)
 
     def get_ikepolicies(self, context, filters=None, fields=None):
-        return self._get_collection(context, vpn_models.IKEPolicy,
-                                    self._make_ikepolicy_dict,
-                                    filters=filters, fields=fields)
+        return model_query.get_collection(context, vpn_models.IKEPolicy,
+                                          self._make_ikepolicy_dict,
+                                          filters=filters, fields=fields)
 
     def _make_ipsecpolicy_dict(self, ipsecpolicy, fields=None):
 
@@ -391,7 +393,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
                'pfs': ipsecpolicy['pfs']
                }
 
-        return self._fields(res, fields)
+        return db_utils.resource_fields(res, fields)
 
     def create_ipsecpolicy(self, context, ipsecpolicy):
         ipsecp = ipsecpolicy['ipsecpolicy']
@@ -400,7 +402,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         lifetime_units = lifetime_info.get('units', 'seconds')
         lifetime_value = lifetime_info.get('value', 3600)
 
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             validator.validate_ipsec_policy(context, ipsecp)
             ipsecp_db = vpn_models.IPsecPolicy(
                 id=uuidutils.generate_uuid(),
@@ -420,7 +422,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
     def update_ipsecpolicy(self, context, ipsecpolicy_id, ipsecpolicy):
         ipsecp = ipsecpolicy['ipsecpolicy']
         validator = self._get_validator()
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             validator.validate_ipsec_policy(context, ipsecp)
             if context.session.query(vpn_models.IPsecSiteConnection).filter_by(
                     ipsecpolicy_id=ipsecpolicy_id).first():
@@ -438,7 +440,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         return self._make_ipsecpolicy_dict(ipsecp_db)
 
     def delete_ipsecpolicy(self, context, ipsecpolicy_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             if context.session.query(vpn_models.IPsecSiteConnection).filter_by(
                     ipsecpolicy_id=ipsecpolicy_id).first():
                 raise vpnaas.IPsecPolicyInUse(ipsecpolicy_id=ipsecpolicy_id)
@@ -452,9 +454,9 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         return self._make_ipsecpolicy_dict(ipsec_db, fields)
 
     def get_ipsecpolicies(self, context, filters=None, fields=None):
-        return self._get_collection(context, vpn_models.IPsecPolicy,
-                                    self._make_ipsecpolicy_dict,
-                                    filters=filters, fields=fields)
+        return model_query.get_collection(context, vpn_models.IPsecPolicy,
+                                          self._make_ipsecpolicy_dict,
+                                          filters=filters, fields=fields)
 
     def _make_vpnservice_dict(self, vpnservice, fields=None):
         res = {'id': vpnservice['id'],
@@ -468,13 +470,13 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
                'external_v4_ip': vpnservice['external_v4_ip'],
                'external_v6_ip': vpnservice['external_v6_ip'],
                'status': vpnservice['status']}
-        return self._fields(res, fields)
+        return db_utils.resource_fields(res, fields)
 
     def create_vpnservice(self, context, vpnservice):
         vpns = vpnservice['vpnservice']
         flavor_id = vpns.get('flavor_id', None)
         validator = self._get_validator()
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             validator.validate_vpnservice(context, vpns)
             vpnservice_db = vpn_models.VPNService(
                 id=uuidutils.generate_uuid(),
@@ -493,7 +495,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
                                 v6_ip=None):
         """Update the external tunnel IP(s) for service."""
         vpns = {'external_v4_ip': v4_ip, 'external_v6_ip': v6_ip}
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             vpns_db = self._get_resource(context, vpn_models.VPNService,
                                          vpnservice_id)
             vpns_db.update(vpns)
@@ -501,7 +503,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
 
     def update_vpnservice(self, context, vpnservice_id, vpnservice):
         vpns = vpnservice['vpnservice']
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             vpns_db = self._get_resource(context, vpn_models.VPNService,
                                          vpnservice_id)
             self.assert_update_allowed(vpns_db)
@@ -510,7 +512,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         return self._make_vpnservice_dict(vpns_db)
 
     def delete_vpnservice(self, context, vpnservice_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             if context.session.query(vpn_models.IPsecSiteConnection).filter_by(
                 vpnservice_id=vpnservice_id
             ).first():
@@ -529,9 +531,9 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         return self._make_vpnservice_dict(vpns_db, fields)
 
     def get_vpnservices(self, context, filters=None, fields=None):
-        return self._get_collection(context, vpn_models.VPNService,
-                                    self._make_vpnservice_dict,
-                                    filters=filters, fields=fields)
+        return model_query.get_collection(context, vpn_models.VPNService,
+                                          self._make_vpnservice_dict,
+                                          filters=filters, fields=fields)
 
     def check_router_in_use(self, context, router_id):
         vpnservices = self.get_vpnservices(
@@ -546,7 +548,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
                                            'services': services})
 
     def check_subnet_in_use(self, context, subnet_id, router_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_READER.using(context):
             vpnservices = context.session.query(
                 vpn_models.VPNService).filter_by(
                     subnet_id=subnet_id, router_id=router_id).first()
@@ -555,8 +557,31 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
                     subnet_id=subnet_id,
                     vpnservice_id=vpnservices['id'])
 
+            query = context.session.query(vpn_models.IPsecSiteConnection)
+            query = query.join(
+                vpn_models.VPNEndpointGroup,
+                vpn_models.VPNEndpointGroup.id ==
+                vpn_models.IPsecSiteConnection.local_ep_group_id).filter(
+                vpn_models.VPNEndpointGroup.endpoint_type ==
+                v_constants.SUBNET_ENDPOINT)
+            query = query.join(
+                vpn_models.VPNEndpoint,
+                vpn_models.VPNEndpoint.endpoint_group_id ==
+                vpn_models.IPsecSiteConnection.local_ep_group_id).filter(
+                vpn_models.VPNEndpoint.endpoint == subnet_id)
+            query = query.join(
+                vpn_models.VPNService,
+                vpn_models.VPNService.id ==
+                vpn_models.IPsecSiteConnection.vpnservice_id).filter(
+                vpn_models.VPNService.router_id == router_id)
+            connection = query.first()
+            if connection:
+                raise vpnaas.SubnetInUseByIPsecSiteConnection(
+                    subnet_id=subnet_id,
+                    ipsec_site_connection_id=connection['id'])
+
     def check_subnet_in_use_by_endpoint_group(self, context, subnet_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_READER.using(context):
             query = context.session.query(vpn_models.VPNEndpointGroup)
             query = query.filter(vpn_models.VPNEndpointGroup.endpoint_type ==
                                  v_constants.SUBNET_ENDPOINT)
@@ -578,12 +603,12 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
                'type': endpoint_group['endpoint_type'],
                'endpoints': [ep['endpoint']
                              for ep in endpoint_group['endpoints']]}
-        return self._fields(res, fields)
+        return db_utils.resource_fields(res, fields)
 
     def create_endpoint_group(self, context, endpoint_group):
         group = endpoint_group['endpoint_group']
         validator = self._get_validator()
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             validator.validate_endpoint_group(context, group)
             endpoint_group_db = vpn_models.VPNEndpointGroup(
                 id=uuidutils.generate_uuid(),
@@ -595,7 +620,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
             for endpoint in group['endpoints']:
                 endpoint_db = vpn_models.VPNEndpoint(
                     endpoint=endpoint,
-                    endpoint_group_id=endpoint_group_db['id']
+                    endpoint_group_id=endpoint_group_db.id
                 )
                 context.session.add(endpoint_db)
         return self._make_endpoint_group_dict(endpoint_group_db)
@@ -604,7 +629,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
                               endpoint_group):
         group_changes = endpoint_group['endpoint_group']
         # Note: Endpoints cannot be changed, so will not do validation
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             endpoint_group_db = self._get_resource(context,
                                                    vpn_models.VPNEndpointGroup,
                                                    endpoint_group_id)
@@ -612,7 +637,7 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         return self._make_endpoint_group_dict(endpoint_group_db)
 
     def delete_endpoint_group(self, context, endpoint_group_id):
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             self.check_endpoint_group_not_in_use(context, endpoint_group_id)
             endpoint_group_db = self._get_resource(
                 context, vpn_models.VPNEndpointGroup, endpoint_group_id)
@@ -624,9 +649,9 @@ class VPNPluginDb(vpnaas.VPNPluginBase,
         return self._make_endpoint_group_dict(endpoint_group_db, fields)
 
     def get_endpoint_groups(self, context, filters=None, fields=None):
-        return self._get_collection(context, vpn_models.VPNEndpointGroup,
-                                    self._make_endpoint_group_dict,
-                                    filters=filters, fields=fields)
+        return model_query.get_collection(context, vpn_models.VPNEndpointGroup,
+                                          self._make_endpoint_group_dict,
+                                          filters=filters, fields=fields)
 
     def check_endpoint_group_not_in_use(self, context, group_id):
         query = context.session.query(vpn_models.IPsecSiteConnection)
@@ -673,7 +698,7 @@ class VPNPluginRpcDbMixin(object):
         The agent will set updated_pending_status as True,
         when agent updates any pending status.
         """
-        with context.session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             for vpnservice in service_status_info_list:
                 try:
                     vpnservice_db = self._get_vpnservice(
@@ -693,15 +718,16 @@ class VPNPluginRpcDbMixin(object):
                         conn['updated_pending_status'])
 
 
-def vpn_callback(resource, event, trigger, **kwargs):
+def vpn_router_gateway_callback(resource, event, trigger, payload=None):
+    # the event payload objects
     vpn_plugin = directory.get_plugin(p_constants.VPN)
     if vpn_plugin:
-        context = kwargs.get('context')
-        router_id = kwargs.get('router_id')
+        context = payload.context
+        router_id = payload.resource_id
         if resource == resources.ROUTER_GATEWAY:
             vpn_plugin.check_router_in_use(context, router_id)
         elif resource == resources.ROUTER_INTERFACE:
-            subnet_id = kwargs.get('subnet_id')
+            subnet_id = payload.metadata.get('subnet_id')
             vpn_plugin.check_subnet_in_use(context, subnet_id, router_id)
 
 
@@ -714,10 +740,10 @@ def migration_callback(resource, event, trigger, **kwargs):
     return True
 
 
-def subnet_callback(resource, event, trigger, **kwargs):
+def subnet_callback(resource, event, trigger, payload=None):
     """Respond to subnet based notifications - see if subnet in use."""
-    context = kwargs['context']
-    subnet_id = kwargs['subnet_id']
+    context = payload.context
+    subnet_id = payload.resource_id
     vpn_plugin = directory.get_plugin(p_constants.VPN)
     if vpn_plugin:
         vpn_plugin.check_subnet_in_use_by_endpoint_group(context, subnet_id)
@@ -725,9 +751,11 @@ def subnet_callback(resource, event, trigger, **kwargs):
 
 def subscribe():
     registry.subscribe(
-        vpn_callback, resources.ROUTER_GATEWAY, events.BEFORE_DELETE)
+        vpn_router_gateway_callback, resources.ROUTER_GATEWAY,
+        events.BEFORE_DELETE)
     registry.subscribe(
-        vpn_callback, resources.ROUTER_INTERFACE, events.BEFORE_DELETE)
+        vpn_router_gateway_callback, resources.ROUTER_INTERFACE,
+        events.BEFORE_DELETE)
     registry.subscribe(
         migration_callback, resources.ROUTER, events.BEFORE_UPDATE)
     registry.subscribe(
